/******************************************************************

      Copyright 1990, 1991, 1992, 1993, by Sun Microsystems, Inc.

Permission to use, copy, modify, distribute, and sell this software
and its documentation for any purpose is hereby granted without fee,
provided that the above copyright notice appear in all copies and
that both that copyright notice and this permission notice appear
in supporting documentation, and that the name of Sun Microsystems, Inc.
not be used in advertising or publicity pertaining to distribution
of the software without specific, written prior permission.
Sun Microsystems, Inc. makes no representations about the suitability of
this software for any purpose.  It is provided "as is" without
express or implied warranty.

Sun Microsystems Inc. DISCLAIMS ALL WARRANTIES WITH REGARD TO THIS
SOFTWARE, INCLUDING ALL IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS,
IN NO EVENT SHALL Sun Microsystems, Inc. BE LIABLE FOR ANY SPECIAL, INDIRECT
OR CONSEQUENTIAL DAMAGES OR ANY DAMAGES WHATSOEVER RESULTING FROM LOSS
OF USE, DATA OR PROFITS, WHETHER IN AN ACTION OF CONTRACT, NEGLIGENCE
OR OTHER TORTIOUS ACTION, ARISING OUT OF OR IN CONNECTION WITH THE USE
OR PERFORMANCE OF THIS SOFTWARE.

  Author: Hideki Hiura (hhiura@Sun.COM)
	  				     Sun Microsystems, Inc.
******************************************************************/

#ifndef lint
#ident  "@(#)XSunIMCanna.c	1.53    93/03/08 SMI"
#endif

#include <stdio.h>
#include <stdlib.h>
#include <fcntl.h>
#ifdef SVR4
#include <sys/utsname.h>
#include <sys/systeminfo.h>
#endif /* SVR4 */
#ifdef X_LOCALE
#include <X11/Xlocale.h>
#else
#include <locale.h>
#endif
#include <sys/types.h>
#include <sys/stat.h>
#include <X11/X.h>
#include <X11/Xlib.h>
#if XlibSpecificationRelease != 5
#include <X11/XlibR5.h>
#endif /* XlibSpecificationRelease */
#include "XSunExt.h"
#include <X11/keysym.h>
#include <X11/Xutil.h>

#ifdef sun
#include <X11/Sunkeysym.h>
#ifndef SunXK_Hangul
#define SunXK_Hangul		0xFF31
#endif
#ifndef SunXK_Hangul_Hanja
#define SunXK_Hangul_Hanja	0xFF34
#endif
#endif

#ifdef SVR4 
#include <libintl.h>
#endif /* SVR4 */

#include <pwd.h>
#include <grp.h>
#include <canna/jrkanji.h>
#include "XSunIMProt.h"
#include "XSunIMPriv.h"
#include "XSunIMMMan.h"
#include "XSunIMMthd.h"
#include "XSunIMCanna.h"
#include "XSunIMPub.h"

#define CANNA_IF_VERSION "1.3"

Public int i18nXView2_COMPAT ;

Private Bool   iml_canna_OpenIF      (/* iml_if_t *If */);
Private Bool   iml_canna_CloseIF     (/* nothing */);
Private Bool   iml_canna_GetIFValue  (/* char *attr, void *value */);
Private Bool   iml_canna_SetIFValue  (/* char *attr, void *value */);
Private void * iml_canna_CreateSC    (/* nothing */);
Private Bool   iml_canna_DestroySC   (/* iml_session_t *s */);
Private Bool   iml_canna_SetSCValue  (/* iml_session_t *s, char *attr, void *value */ );
Private Bool   iml_canna_GetSCValue  (/* iml_session_t *s, char *attr, void *value */ );
Private char * iml_canna_ResetSC     (/* iml_session_t *s */);
Private void   iml_canna_SetSCFocus  (/* iml_session_t *s */);
Private void   iml_canna_UnsetSCFocus(/* iml_session_t *s */);
Private void   iml_canna_SendEvent   (/* iml_session_t *s , XEvent *e */);
Private Bool   iml_canna_FilterEvent (/* XEvent *e */);

#ifdef DYNAMICIF
Private
#else
Public
#endif  
if_methods_t canna_methods = {
    iml_canna_OpenIF,      
    iml_canna_CloseIF,     
    iml_canna_GetIFValue,  
    iml_canna_SetIFValue,  
    iml_canna_CreateSC,    
    iml_canna_DestroySC,   
    iml_canna_SetSCValue,  
    iml_canna_GetSCValue,  
    iml_canna_ResetSC,     
    iml_canna_SetSCFocus,  
    iml_canna_UnsetSCFocus,
    iml_canna_SendEvent,
    iml_canna_FilterEvent,
} ;

Private Bool
iml_canna_OpenIF(If)
iml_if_t *If ;
{
    return True ;
}
Private Bool
iml_canna_CloseIF(If)
iml_if_t *If ;
{
    return True;
}
Private Bool
iml_canna_GetIFValue(If, attr, value)
iml_if_t *If ;
char *attr;
void *value;
{
    if(!strcmp(attr, IF_VERSION)){
	*(char**)value = CANNA_IF_VERSION ;
	return True ;
    }
    return False ;
}
Private Bool
iml_canna_SetIFValue(If, attr, value)
iml_if_t *If ;
char *attr;
void *value;
{
    return False ;
}
Private void * 
iml_canna_CreateSC(If, ic, le_name)
iml_if_t *If ;
void *ic ;
char *le_name ; /* This is for multiple language engine per I/F */
{
    Private int canna_create_session(/*int id,iml_session_t *s,Bool is_recycled*/);
    int le_id;
    iml_session_t *s ;

    if(le_name == NULL){
	le_id = 0 ;
    } else {
	le_id = atoi(le_name);
    }
    return (void *) If->m->iml_construct_session(If, ic, le_id,sizeof(iml_session_t),canna_create_session);
}
Private Bool
iml_canna_DestroySC(s)
iml_session_t *s ;
{
    canna_destruct_session(s);
}
Private char * 
iml_canna_ResetSC(s)
iml_session_t *s ;
{
    canna_send_commit(s);
}

Private Bool
iml_canna_SetSCValue(s, attr, value)
iml_session_t *s ;
char *attr ;
void *value ;
{
    if(!strcmp(attr, SC_Henkan_Mode) || !strcmp(attr, XNExtXimp_Conversion)){
	if((Bool)value == True){
	    if(!(s->public_status & IMLSTATUS_Henkan_Mode)){
		s->public_status |= IMLSTATUS_Henkan_Mode ;
		canna_make_conversion_on(s);
	    }
	} else {
	    if(s->public_status & IMLSTATUS_Henkan_Mode){
		s->public_status &= ~IMLSTATUS_Henkan_Mode ;
		canna_make_conversion_off(s);
	    }
	}
	return True ;
    } else if(!strcmp(attr, SC_STATUS)){
	if(!(s->public_status & IMLSTATUS_Henkan_Mode)){
	    if((iml_status_t)value & IMLSTATUS_Henkan_Mode){
		canna_make_conversion_on(s);
	    }
	}
	s->public_status = (iml_status_t)value ;
	return True ;
    } 
}

Private Bool
iml_canna_GetSCValue(s, attr, value)
iml_session_t *s ;
char *attr ;
void *value ;
{
    if(!strcmp(attr, SC_STATUS)){
	iml_inst *lp = s->If->m->iml_make_status_notify_inst(s);
	*(iml_status_t*)value = s->public_status ;
	if(lp = s->If->m->iml_execute(s, &lp)){
	    canna_eval_results(s, lp);
	}
	return True ;
    }
    return False ;
    
}
Private void
iml_canna_SetSCFocus(s)
iml_session_t *s ;
{
    iml_inst *lp = s->If->m->iml_make_status_notify_inst(s);
    if(lp = s->If->m->iml_execute(s, &lp)){
	canna_eval_results(s, lp);
    }
}
Private void
iml_canna_UnsetSCFocus(s)
iml_session_t *s ;
{
    iml_inst *lp = s->If->m->iml_make_status_notify_inst(s);
    if(lp = s->If->m->iml_execute(s, &lp)){
	canna_eval_results(s, lp);
    }
}
    
Private Bool translate_keysym_to_canna_keycode(/*s, keysym, canna_keycode*/) ;

#define canna_is_conv_on_key(s,l) ((l)>0 && ((s)->conversion_on_key == *((s)->XLookupBuf)))
Private void
iml_canna_SendEvent(s, e)
iml_session_t *s ;
XKeyEvent *e;  /* If this is NULL, this event must be pseudo event */
{
    int l ; /* length of Lookuped string */
    iml_inst *lp ;
    unsigned int   save_state;
    
    /*
     * canna engine $@$b(J KeyPress Event $@$7$+$$$i$J$$$@$m$&!#(J
     */
    if(e && e->type != KeyPress){
	return ; 
    }
    if(e){
	s->event = e ; /* save current KeyEvent */
    } else {
	e = s->event ; /* May NOT be needed: restore KeyEvent */
    }
    save_state = e->state;
    if(s->public_status & IMLSTATUS_Henkan_Mode && e->state & ControlMask) 
	e->state &= ~e->display->mode_switch;
    l= s->If->m->iml_lookupString(e,s->XLookupBuf,s->XLookupBuflen,&s->keysym);
    e->state = save_state;

#ifdef notdef
    /*
     * TYPE1 $@MQ$N=hM}$,<BAu$5$l$k$^$G!"$*$"$:$1!#(J
     * $@8=:_$O!"(JXIMP_TYPE $@$r(J TYPE2 $@$K@_Dj$9$k$H$$$&@)8B;v9`$H(J
     * $@$H$7$F$*$/(J
     */
    if(! (s->public_status & IMLSTATUS_Henkan_Mode)){
	/*
	 * $@L5JQ49$N>l9g(J
	 */
	iml_inst *rv ;
	if((s->keysym == XK_Henkan_Mode) || 
	   (s->keysym == SunXK_Hangul) || canna_is_conv_on_key(s,l)){
	    canna_make_conversion_on(s);
	    return ;
	} 
	lp = s->If->m->iml_make_keypress_inst(s, s->XLookupBuf) ;
	if(lp = s->If->m->iml_execute(s, &lp)){
	    canna_eval_results(s, lp);
	}
	return ;
    }
#endif    
    /*
     * $@JQ49(J ON $@$N;~(J
     */
    if(translate_keysym_to_canna_keycode(s, l)){ /* data would be stored in s */
	iml_inst **rrv ;
	s->cslen = jrKanjiString(s->session_id, s->XLookupBuf[0],
					s->csbuf, s->csbuflen, &s->ks);
	eval_canna_to_env_packet(s);
	return;
    } else {
	/*
	 * canna $@$KAw$l$J$$$N$G$O;EJ}$,$"$k$^$$!#(J
	 * $@2?$,5/$-$k$+CN$i$s$,!"(Jkeypress $@$rAw$C$F$7$^$(!#(J
	 */
	lp = s->If->m->iml_make_keypress_inst(s,s->XLookupBuf) ;
	if(lp = s->If->m->iml_execute(s, &lp)){
	    cm_eval_results(s, lp);
	}
    }
    return;
}

Private Bool
iml_canna_FilterEvent(e)
XEvent *e ;
{
    return False ;
}

Private void
canna_conversion_on(s)
iml_session_t *s;
{
    iml_inst *rv ;

    if(s->remainder){
	rv = s->remainder ;
	s->remainder = NULL ;
	return;
    }
    canna_make_conversion_on(s);
}

Private void
canna_eval_results(s, results)
iml_session_t *s;
iml_inst *results ;
{
    XIMLookupStartCallbackStruct *ls;
    XIMLookupProcessCallbackStruct *lp;
    int *i ;
    
    if(results->opcode & IMM_CB_RESULT_REQUIRED){
	results->opcode &= ~IMM_CB_RESULT_REQUIRED ;
	switch(results->opcode){
	  case IMM_LOOKUP_START:
	    ls = (XIMLookupStartCallbackStruct *)& results->operand ;
	    if(s->WhoIsMaster == HasNotBeenNegotiated) {
	      s->WhoIsMaster =  ls->WhoIsMaster ;
	    } else {
		break;
	    }
	    /*
	     * Obey CBPreferrence
	     */
	    s->XIMPreference = s->CBPreference ;
	    break;
	  case IMM_LOOKUP_PROCESS:
	    lp = (XIMLookupProcessCallbackStruct *)& results->operand ;
            switch(lp->index_of_choice_selected) {
                case XIM_UNDETERMINED:
                case XIM_UNKNOWN_KEYSYM:
                  /*
                   * invalid label input is done.
                   */
                        break;
                case XIM_UNKNOWN_CONTROL:
                        if(s->WhoIsMaster == CBIsMaster)
                                canna_select_unknown_control(s);
                        break;
                default:
                        canna_set_choice(s, lp->index_of_choice_selected);
            }
	    return;
	case IMM_PREEDIT_START:
	    s->PreEditBufferSizeInCallback = *(int *)&results->operand ;
	    break;
	  default:
	    break;
	}
	return ;
    }
}

Private void
canna_make_conversion_off(s)
iml_session_t *s ;
{
    canna_send_commit(s);
}
Private void
canna_make_conversion_on(s)
iml_session_t *s ;
{
    iml_inst *rv = NULL;
    iml_inst *lp ;
    
    s->public_status |= IMLSTATUS_Henkan_Mode ;
    s->ksv.val = CANNA_MODE_HenkanMode ;
    jrKanjiControl(s->session_id, KC_CHANGEMODE, (void*)&s->ksv);
    
    lp = s->If->m->iml_make_status_notify_inst(s);
    s->If->m->iml_link_inst_tail(&rv, lp);
    lp = s->If->m->iml_make_preedit_start_inst(s);
    s->If->m->iml_link_inst_tail(&rv, lp);
    if(lp = s->If->m->iml_execute(s, &rv)){
	canna_eval_results(s, lp);
    }
    eval_canna_to_env_packet(s);
}
Private
canna_create_session(canna_id, s, is_recycled)
int canna_id ;
iml_session_t *s ;
Bool is_recycled ;
{
    char **warning ;
    int session_id = (int)s ; /* caller is responsible to make session_id in canna */

    if(jrKanjiControl(session_id, KC_INITIALIZE, (void*)&warning) == -1){
        fprintf(stderr,dgettext("SUNW_XIM_HTT_MSG",
				"canna I/F: language engine open failed.\n"));
	return(-1);
    }
    if (warning) {
	char **p;

	for (p = warning ; *p ; p++) {
	    fprintf(stderr, "%s0", *p);
        }
    }
    if (is_recycled) {
	bzero(s->csbuf,s->csbuflen);
    } else {
	s->csbuf = (char *) calloc(sizeof(char), DEFAULTCommitedStringBufferSize);
	s->csbuflen = DEFAULTCommitedStringBufferSize ;
	s->ksv.buffer = s->csbuf ;
	s->ksv.bytes_buffer = s->csbuflen ;
	s->ksv.ks = &s->ks ;
    }
    s->cslen = 0;
    s->ksv.val = 0 ;
    bzero(&s->ks, sizeof(s->ks));
    return(session_id);
}

Private 
env_commit(s, rrv)
iml_session_t *s;
iml_inst **rrv;
{
    iml_inst *lp ;
    
    if(i18nXView2_COMPAT){
	lp = s->If->m->iml_make_preedit_erase_inst(s);
	s->If->m->iml_link_inst_tail(rrv, lp);
	lp = s->If->m->iml_make_preedit_done_inst(s);
	s->If->m->iml_link_inst_tail(rrv, lp);
    }    
    lp = s->If->m->iml_make_commit_inst(s, s->cslen, s->csbuf);
    s->If->m->iml_link_inst_tail(rrv, lp);

}

#define canna_attr2feedback(p, revPos, revEnd, len)\
{                                                  \
    register i;                                    \
    XIMFeedback *cfb = (p);                        \
                                                   \
    for(i=0;i<len;i++){                            \
	cfb[i] = 0 ;       /* default feedback */  \
    }                                              \
    for(i=(revPos);i<(revEnd);i++){                \
	cfb[i] |= XIMReverse ;                     \
    }                                              \
}

Private iml_inst *
make_preedit_draw_inst(s)
iml_session_t *s;
{
    char    *cmbs = s->ks.echoStr ;                         /* current mbstr */
    int      cmbl = s->ks.length;                           /* current mblen */
    wchar_t *cws ;                                          /* current wstr */
    int      cwl ;                                          /* current len */
    XIMFeedback *cfb ;                                      /* current fb */
    int      endPos = s->ks.revPos + s->ks.revLen ;

    if(cmbs && cmbl > 0){
	cws = (wchar_t*)_iml_new(s, (cmbl+1)*sizeof(wchar_t));
	mbstowcs(cws,cmbs,IML_MAX_SLOT_SIZE);
	cwl = wslen(cws);
	cfb = _iml_new(s, (cmbl + 1)*sizeof(XIMFeedback));
	canna_attr2feedback(cfb, s->ks.revPos, endPos, cwl);
	return s->If->m->iml_make_preedit_draw_inst(s, cws, cfb);
    } else {
	return s->If->m->iml_make_preedit_erase_inst(s);
    }
}

Private 
canna_interm(s, rrv)
iml_session_t *s;
iml_inst **rrv;
{
    iml_inst *lp ;
    
    if(!IS_REGION_ACTIVE(s, PREEDIT)){
	ACTIVATE_REGION(s, PREEDIT) ;
	lp = s->If->m->iml_make_preedit_start_inst(s);
	if(lp = s->If->m->iml_execute(s, &lp)){
	    canna_eval_results(s, lp);
	}
    }
    lp = make_preedit_draw_inst(s);
    s->If->m->iml_link_inst_tail(rrv, lp);
    _imm_trace("IMM_SET_REGION(pseudo PREEDIT)",0,0);
}
Private  void
canna_update_footer(s, rrv, is_footer_lookup)
iml_session_t *s;
iml_inst **rrv;
Bool is_footer_lookup;
{
    iml_inst *lp ;
    wchar_t *ws = s->status_cache.ws ;
    register i, l ;
    XIMFeedback *fb ;
    
    if(!IS_REGION_ACTIVE(s, STATUS)){
	ACTIVATE_REGION(s, STATUS) ;
	lp = s->If->m->iml_make_status_start_inst(s);
	s->If->m->iml_link_inst_tail(rrv, lp);
    }
    if(is_footer_lookup){
	int endPos = s->ks.gline.revPos + s->ks.gline.revLen ;
	if(s->ks.gline.line && s->ks.gline.length > 0){
	    l = mbstowcs(ws+s->status_cache.len+1, s->ks.gline.line,s->status_cache.size);
	    canna_attr2feedback(s->status_cache.fb+s->status_cache.len+1,
				s->ks.gline.revPos, endPos, l);
	    lp = s->If->m->iml_make_status_draw_inst(s, ws, fb) ;
	}
    } else {
	l = s->status_cache.len = mbstowcs(ws, s->ks.mode,s->status_cache.size);
	for(i = 0 ; i < l ; i++)
	    s->status_cache.sfb[i] = s->status_cache.fb[i] | XIMTertiary ; /* for stipple */
	if(s->private_status & BEING_FOCUSED)
	    lp = s->If->m->iml_make_status_draw_inst(s, ws, s->status_cache.fb) ;
	else
	    lp = s->If->m->iml_make_status_draw_inst(s, ws, s->status_cache.sfb) ;
    }
    s->If->m->iml_link_inst_tail(rrv, lp);
}

Private iml_inst **
eval_it(s, rrv)
iml_session_t *s ;
iml_inst **rrv;
{
    iml_inst *lp = 0;
    int op ; 

    if(s->cslen != -1){
	canna_interm(s, rrv); /* preedit region update */
	s->If->m->iml_link_inst_tail(rrv, lp);
    }
    if(s->ks.info & KanjiModeInfo){
	canna_update_footer(s, rrv, False);
	s->If->m->iml_link_inst_tail(rrv, lp);
    }
    if(s->ks.info & KanjiGLineInfo){
	canna_update_footer(s, rrv, True);
	s->If->m->iml_link_inst_tail(rrv, lp);
    }
    return(&lp->next); /* iml_inst lin last point */
}

Private void
eval_canna_to_env_packet(s)
iml_session_t *s ;
{
    iml_inst *rv = NULL ;
    eval_it(s, &rv);
    if(rv = s->If->m->iml_execute(s, &rv)){
	canna_eval_results(s, rv);
    }
}
Private void
canna_destruct_session(s)
iml_session_t *s ;
{
    Private void destruct_session(/* iml_session_t *s */);
    s->If->m->iml_destruct_session(s, destruct_session);

}
Private void
destruct_session(s)
iml_session_t *s ;
{
    char **warning;
    
    jrKanjiControl(s->session_id, KC_FINALIZE, (void *)&warning);
    if (warning) {
	char **p;
	
	for (p = warning ; *p ; p++) {
	    fprintf(stderr, "%s0", *p);
	}
    }
    eval_canna_to_env_packet(s);
    return;
}

Private void
canna_reset(s, rrv)
iml_session_t *s ;
iml_inst **rrv ;
{
    iml_inst *lp ;
    lp = s->If->m->iml_make_nop_inst(s);
    s->If->m->iml_link_inst_tail(rrv, lp);
}
Private void
canna_send_commit(s)
iml_session_t *s;
{
    if(s->remainder){
	if(s->remainder = s->If->m->iml_execute(s, &s->remainder)){
	    canna_eval_results(s, s->remainder);
	}
        s->remainder = NULL;
        return;
    }
    jrKanjiControl(s->session_id, KC_KAKUTEI, (void *)&s->ksv);
    eval_canna_to_env_packet(s);
}

Private void
canna_set_choice(s, i)
iml_session_t *s ;
int i ; /* index */
{
}

Private void
canna_select_unknown_control(s)
iml_session_t *s ;
{
}

Private Bool
translate_keysym_to_canna_keycode(s, keysym)
iml_session_t *s ;
KeySym keysym ;
{
#define functionKey(keysym,s) 
    switch(keysym){
      case XK_Shift_L:
      case XK_Shift_R:
      case XK_Control_L:
      case XK_Control_R:
      case XK_Caps_Lock:
      case XK_Shift_Lock:
      case XK_Meta_L:
      case XK_Meta_R:
      case XK_Alt_L:
      case XK_Alt_R:
      case XK_Super_L:
      case XK_Super_R:
      case XK_Hyper_L:
      case XK_Hyper_R:
	/*
	 * Only modifier key is coming..
	 */
	return False;
	/*
	 * Function Key Operations
	 */
      case XK_Multi_key: /* 0xFF20  Multi-key character compose */
	break;
      case XK_Kanji: /* 0xFF21	Kanji, Kanji convert */
	s->XLookupBuf[0] = CANNA_KEY_Xfer ;
	goto VALID;
	break;
      case XK_Muhenkan: /* 0xFF22   Cancel Conversion */
	break;
      case SunXK_Hangul: /* 0xFF31  Start/Stop Conversion */
      case XK_Henkan_Mode: /* 0xFF23  Start/Stop Conversion */
	s->XLookupBuf[0] = CANNA_KEY_Shift_Nfer ;
	goto VALID;
	break;
      case XK_Romaji: /* 0xFF24   to Romaji */
	break;
      case XK_Hiragana: /* 0xFF25   to Hiragana */
	break;
      case XK_Katakana: /* 0xFF26   to Katakana */
	break;
      case XK_Hiragana_Katakana: /* 0xFF27  Hiragana/Katakana toggle */
	break;
      case XK_Zenkaku: /* 0xFF28   to Zenkaku */
	break;
      case XK_Hankaku: /* 0xFF29   to Hankaku */
	break;
      case XK_Zenkaku_Hankaku: /* 0xFF2A   Zenkaku/Hankaku toggle */
	break;
      case XK_Touroku: /* 0xFF2B   Add to Dictionary */
	break;
      case XK_Massyo: /* 0xFF2C   Delete from Dictionary */
	break;
      case XK_Kana_Lock: /* 0xFF2D  Kana Lock */
	break;
      case XK_Kana_Shift: /* 0xFF2E  Kana Shift */
	break;
      case XK_Eisu_Shift: /* 0xFF2F  Alphanumeric Shift */
	break;
      case XK_Eisu_toggle: /* 0xFF30  Alphanumeric toggle */
	break;
      case XK_Home: /* 	0xFF50 */
	break;
      case XK_Left: /* 	0xFF51 Move left, left arrow */
	break;
      case XK_Up: /* 	0xFF52	/* Move up, up arrow */
	break;
      case XK_Right: /* 0xFF53	/* Move right, right arrow */
	break;
      case XK_Down: /* 	0xFF54	/* Move down, down arrow */
	break;
      case XK_Prior: /* 0xFF55	/* Prior, previous */
	break;
      case XK_Next: /* 	0xFF56	/* Next */
	break;
      case XK_End: /* 	0xFF57	/* EOL */
	break;
      case XK_Begin: /* 0xFF58	/* BOL */
	break;
      case XK_Select: /* 0xFF60	 Select, mark */
	break;
      case XK_Print: /* 0xFF61 */
	break;
      case XK_Execute: /* 0xFF62  Execute, run, do */
	s->XLookupBuf[0] = CANNA_KEY_Xfer ;
	goto VALID;
	break;
      case XK_Insert: /* 0xFF63	 Insert, insert here */
	break;
      case XK_Undo: /* 	0xFF65	Undo, oops */
	break;
      case XK_Redo: /* 	0xFF66	redo, again */
	break;
      case XK_Menu: /* 	0xFF67 */
	break;
      case XK_Find: /* 	0xFF68	 Find, search */
	break;
      case XK_Cancel: /* 0xFF69	Cancel, stop, abort, exit */
	break;
      case XK_Help: /* 	0xFF6A	Help, ? */
	break;
      case XK_Break: /* 0xFF6B */
	break;
      case XK_Mode_switch: /* 0xFF7E Character set switch */
	/* case XK_script_switch: 0xFF7E Alias for mode_switch */
	/*
	 * $@$3$l$GJQ$K$J$k$N$O!"(Jcanna $@$,$o$k$$!#(J
	 * XIM $@$N(J bug $@$K$5$l$k$+$i!"$3$3$GBP=h$9$k$N$O$d$a$k!#(J
	 * canna $@$rD>$9$Y$-(J
	 */
    	s->XLookupBuf[0] = CANNA_KEY_Home;
    	goto VALID;
	break;
#ifdef sun
      case SunXK_Hangul_Hanja: /* 0xFF34 HangleCharacter set switch */
    	s->XLookupBuf[0] = CANNA_KEY_Home ;
    	goto VALID;
	break;
#endif
      case XK_Num_Lock: /* 0xFF7F */
	break;
      case XK_KP_Space: /* 0xFF80	 space */
	break;
      case XK_KP_Tab: /* 0xFF89 */
	break;
      case XK_KP_Enter: /* 0xFF8D enter */
	break;
      case XK_KP_F1: /* 0xFF91	PF1, KP_A, ... */
	break;
      case XK_KP_F2: /* 0xFF92 */
	break;
      case XK_KP_F3: /* 0xFF93 */
	break;
      case XK_KP_F4: /* 0xFF94 */
	break;
      case XK_KP_Equal: /* 0xFFBD  equals */
	s->XLookupBuf[0] = '=';
	goto VALID;
	break;
      case XK_KP_Multiply: /* 0xFFAA */
	s->XLookupBuf[0] = '*';
	goto VALID;
	break;
      case XK_KP_Add: /* 0xFFAB */
	s->XLookupBuf[0] = '+';
	goto VALID;
	break;
      case XK_KP_Separator: /* 0xFFAC separator, often comma */
	s->XLookupBuf[0] = ',';
	goto VALID;
	break;
      case XK_KP_Subtract: /* 0xFFAD */
	s->XLookupBuf[0] = '-';
	goto VALID;
	break;
      case XK_KP_Decimal: /* 0xFFAE */
	s->XLookupBuf[0] = '.';
	goto VALID;
	break;
      case XK_KP_Divide: /* 0xFFAF */
	s->XLookupBuf[0] = '/';
	goto VALID;
	break;
      case XK_KP_0: /* 	0xFFB0 */
	s->XLookupBuf[0] = '0';
	goto VALID;
	break;
      case XK_KP_1: /* 	0xFFB1 */
	s->XLookupBuf[0] = '1';
	goto VALID;
	break;
      case XK_KP_2: /* 	0xFFB2 */
	s->XLookupBuf[0] = '2';
	goto VALID;
	break;
      case XK_KP_3: /* 	0xFFB3 */
	s->XLookupBuf[0] = '3';
	goto VALID;
	break;
      case XK_KP_4: /* 	0xFFB4 */
	s->XLookupBuf[0] = '4';
	goto VALID;
	break;
      case XK_KP_5: /* 	0xFFB5 */
	s->XLookupBuf[0] = '5';
	goto VALID;
	break;
      case XK_KP_6: /* 	0xFFB6 */
	s->XLookupBuf[0] = '6';
	goto VALID;
	break;
      case XK_KP_7: /* 	0xFFB7 */
	s->XLookupBuf[0] = '7';
	goto VALID;
	break;
      case XK_KP_8: /* 	0xFFB8 */
	s->XLookupBuf[0] = '8';
	goto VALID;
	break;
      case XK_KP_9: /*  0xFFB9 */
	s->XLookupBuf[0] = '9';
	goto VALID;
	break;
      case XK_F1:
	s->XLookupBuf[0] = CANNA_KEY_F1 ;
	goto VALID;
	break;
      case XK_F2:
	s->XLookupBuf[0] = CANNA_KEY_Xfer ;
	goto VALID;
	break;
      case XK_F3:
	break;
      case XK_F4:
	break;
      case XK_F5:
	break;
      case XK_F6:
	break;
      case XK_F7:
	break;
      case XK_F8:
	break;
      case XK_F9:
	break;
      case XK_F10:
	break;
#if XlibSpecificationRelease != 5
#ifdef OWNV2
      case XK_SunF36: 
#endif /* sun */
#endif /* XlibSpecificationRelease */
      case SunXK_F36: /* Labeled F11 */
	break;
#if XlibSpecificationRelease != 5
#ifdef OWNV2
      case XK_SunF37: 
#endif /* sun */
#endif /* XlibSpecificationRelease */
      case SunXK_F37: /* Labeled F12 */
	break;
/*
 * Left/right function key $@$r(J trap $@$9$k$N$O$b$&$d$a$?!#(J
 */
#define break /**/
      case XK_F11: /* XK_L1 */
	break;
      case XK_F12: /* XK_L2 */
	break;
      case XK_F13: /* XK_L3 */
	break;
      case XK_F14: /* XK_L4 */
	break;
      case XK_F15: /* XK_L5 */
	break;
      case XK_F16: /* XK_L6 */
	break;
      case XK_F17: /* XK_L7 */
	break;
      case XK_F18: /* XK_L8 */
	break;
      case XK_F19: /* XK_L9 */
	break;
      case XK_F20: /* XK_L10 */
	break;
      case XK_F21: /* XK_R1 */
	break;
      case XK_F22: /* XK_R2 */
	break;
      case XK_F23: /* XK_R3 */
	break;
      case XK_F24: /* XK_R4 */
	break;
      case XK_F25: /* XK_R5 */
	break;
      case XK_F26: /* XK_R6 */
	break;
      case XK_F27: /* XK_R7 */
	break;
      case XK_F28: /* XK_R8 */
	break;
      case XK_F29: /* XK_R9 */
	break;
      case XK_F30: /* XK_R10 */
	break;
      case XK_F31: /* XK_R11 */
	break;
      case XK_F32: /* XK_R12 */
	break;
      case XK_R13: /* XK_F33 */
	break;
      case XK_F34: /* XK_R14 */
	break;
      case XK_F35: /* XK_R15 */
	break;
	return False ;
#undef break
      default:	
VALID:
	return True ;
    }
    return False ;
}
